"""
misgnomer.py
by Daniel Bickett
dbickett@gmail.com
http://www.heureusement.org/

FINAL RELEASE
I got home with too little time to polish it and 
turn it into an executable, so this will have to
do :(
"""

import pygame
pygame.init()

import random
import Queue

###########################
##                         
##  PRELIMINARY DATA/CODE! 
##                         

# Constants
# - Directional
DIR_N = 101
DIR_E = 102
DIR_S = 103
DIR_W = 104
# - Loop codes
LP_DEAD = 201
LP_WON = 202
LP_TIMEOUT = 203

# Centralized Data: 'alloc' dict
# - Will Contain Graphics/Objects
# - Helps differentiate between locals and globals;
#   aids in orientation.

# screen related data
alloc = {}
alloc['screen_res'] = 608, 540
alloc['screen'] = pygame.display.set_mode( alloc['screen_res'] )

# load/render fonts
alloc['font_color'] = 255, 181, 140
alloc['font_large'] = pygame.font.Font( "BENGUIAB.TTF" , 36 )
alloc['font_medium'] = pygame.font.Font( "BENGUIAB.TTF" , 24 )
alloc['pause'] = alloc['font_large'].render( "PAUSE" , True , alloc['font_color'] )
alloc['pause_rect'] = alloc['pause'].get_rect()
alloc['death'] = alloc['font_large'].render( "Death" , True , ( 255 , 255 , 255 ) )
alloc['death_rect'] = alloc['death'].get_rect()
alloc['success'] = alloc['font_large'].render( "Success" , True , ( 255 , 255 , 255 ) )
alloc['success_rect'] = alloc['success'].get_rect()
alloc['success_caption'] = alloc['font_medium'].render( "Another gnome returned home" , True , ( 255 , 255 , 255 ) )
alloc['success_caption_rect'] = alloc['success_caption'].get_rect()

# load the images
alloc['image'] = {}
alloc['image']['background'] = pygame.image.load("background.gif").convert()
alloc['image']['cloud'] = pygame.image.load("cloud.gif").convert()
alloc['image']['shadow'] = pygame.image.load("cloud_shadow.gif").convert()
alloc['image']['shadow'].set_alpha( 130 )
alloc['image']['hero'] = pygame.image.load("hero.gif").convert()
alloc['image']['life_container'] = pygame.image.load("life_container.gif").convert()
alloc['image']['life_unit'] = pygame.image.load("life_unit.gif").convert()
alloc['image']['tree'] = pygame.image.load("tree.gif").convert()
alloc['image']['tree_shadow'] = pygame.image.load("tree_shadow.gif").convert()
alloc['image']['tree_shadow'].set_alpha( 130 )

for image in alloc['image'].itervalues():
    image.set_colorkey( ( 255 , 140 , 237 ) )

# rect objects
alloc['cloud_rect'] = alloc['image']['cloud'].get_rect()
alloc['ground_rect'] = pygame.Rect( 0 , 123 , 608 , 417 )
alloc['screen_rect'] = pygame.Rect( ( 0 , 0 ) , alloc['screen_res'] )
alloc['sky_rect'] = pygame.Rect( 0 , 0 , 608 , 123 )
alloc['hero_rect'] = alloc['image']['hero'].get_rect()
alloc['win_rect'] = pygame.Rect( 314 , 123 , 30 , 5 )

def centerImgOnScreen( image , rect ):
    """
    Give this an image and its rect, and it will return
    a topleft x,y coord for the image, placing it in the
    center of the screen.
    """
    x = ( alloc['screen_rect'].right / 2 ) - ( rect.right / 2 )
    y = ( alloc['screen_rect'].bottom / 2 ) - ( rect.bottom / 2 )
    return x,y

##########################
##                        
##  VISIBLE GAME OBJECTS! 
##                        

class _Shadow:

    def __init__( self , initpos ):
        self.image = alloc['image']['shadow']
        self._setPos( initpos )

    def _setPos( self , pos ):
        self.rect = self.image.get_rect().move( pos )

class _Cloud:

    def __init__( self , initpos , speed , direction , shadowSpaceRect , shadKey ):
        self.image = alloc['image']['cloud']
        shadowSpaceRect.left = initpos[0]
        self.shadow = _Shadow( shadowSpaceRect.topleft )
        alloc['shadows'][ tuple( self.shadow.rect ) ] = 1
        self._setPos( initpos )
        self.speed = speed
        self.direction = direction

    def _setPos( self , pos ):
        self.rect = self.image.get_rect().move( pos )
        self.shadow._setPos( ( pos[0] , self.shadow.rect.top ) )

    def _step( self ):
        # check to see if this cloud needs to be recycled
        if ( self.direction == DIR_E and self.rect.left > alloc['screen_rect'].right ) or \
        ( self.direction == DIR_W and self.rect.right < 0 ):
            return False
        else:
            # step in the proper direction at the proper speed
            if self.direction == DIR_E:
                self.rect.right += self.speed
                self.shadow.rect.right += self.speed
            elif self.direction == DIR_W:
                self.rect.left -= self.speed
                self.shadow.rect.left -= self.speed
            alloc['shadows'][ tuple( self.shadow.rect ) ] = 1
            return True

    def _blit( self ):
        alloc['screen'].blit( self.image , self.rect )
        alloc['screen'].blit( self.shadow.image , self.shadow.rect )

    def next( self ):
        if self._step():
            self._blit()
            return True
        else:
            return False

class Clouds:
    
    def __init__( self , min_speed = 1 , max_speed = 3 ):
        # NOTE: The cloud creation used to be done in a loop, but now we
        #       need to be more precise on account of the shadows. I'm
        #       fairly sure each level will only have four clouds, so I
        #       don't think this will become a problem.
        self.min_speed = min_speed
        self.max_speed = max_speed
        # work with shadows
        shadowSpaceHeight = int( ( alloc['ground_rect'].bottom - 128 ) / 7 )
        self.shadspc = {}
        self.shadspc['A'] = pygame.Rect( alloc['ground_rect'].topleft , ( 608 , shadowSpaceHeight ) )
        self.shadspc['B'] = pygame.Rect( alloc['ground_rect'].left , self.shadspc['A'].top + shadowSpaceHeight , 608 , shadowSpaceHeight )
        self.shadspc['C'] = pygame.Rect( alloc['ground_rect'].left , self.shadspc['B'].top + shadowSpaceHeight , 608 , shadowSpaceHeight )
        self.shadspc['D'] = pygame.Rect( alloc['ground_rect'].left , self.shadspc['C'].top + shadowSpaceHeight , 608 , shadowSpaceHeight )
        self.shadspc['E'] = pygame.Rect( alloc['ground_rect'].left , self.shadspc['D'].top + shadowSpaceHeight , 608 , shadowSpaceHeight )
        # clouds' container
        self.clouds = {}
        self.clouds['A'] = self._genCloud( self.shadspc['A'] , DIR_W , 'A' )
        self.clouds['B'] = self._genCloud( self.shadspc['B'] , DIR_E , 'B' )
        self.clouds['C'] = self._genCloud( self.shadspc['C'] , DIR_W , 'C' )
        self.clouds['D'] = self._genCloud( self.shadspc['D'] , DIR_E , 'D' )
        self.clouds['E'] = self._genCloud( self.shadspc['E'] , DIR_W , 'E' )

    def _genCred( self , direction ):
        # position
        coord = [0,0]
        if direction == DIR_E:
            coord[0] = - alloc['cloud_rect'].right
        else:
            coord[0] = alloc['sky_rect'].right + alloc['cloud_rect'].right
        coord[1] = random.randint( alloc['sky_rect'].top , alloc['sky_rect'].bottom - alloc['cloud_rect'].bottom )
        # speed
        speed = random.randint( self.min_speed , self.max_speed )
        # done generating.
        return speed, coord
    
    def _genCloud( self , shadowSpace , direction , shadKey ):
        speed, coord = self._genCred( direction )
        return _Cloud( coord , speed , direction , shadowSpace , shadKey )

    def next( self ):
        for cloud in self.clouds.itervalues():
            if not cloud.next():
                speed, coord = self._genCred( cloud.direction )
                cloud._setPos( coord )
                cloud.speed = speed

class Tree:
    
    def __init__( self ):
        self.image = alloc['image']['tree']
        self.pos = self.image.get_rect()
        self.rect = pygame.Rect( ( ( alloc['screen_rect'].right / 2 ) - ( self.pos.right / 2 ) + 50 , alloc['screen_rect'].bottom - self.pos.bottom ) , self.pos.bottomright )
        self.shadow_image = alloc['image']['tree_shadow']
        self.shadow_pos = self.shadow_image.get_rect()
        self.shadow_rect = pygame.Rect( ( ( alloc['screen_rect'].right / 2 ) - ( self.shadow_pos.right / 2 ) + 50 , alloc['screen_rect'].bottom - self.shadow_pos.bottom ) , self.shadow_pos.bottomright )
        alloc['shadows'][ tuple( self.shadow_rect ) ] = 1
        alloc['shadows'][ tuple( self.rect ) ] = 1

    def blit( self ):
        alloc['screen'].blit( self.shadow_image , self.shadow_rect )
        alloc['screen'].blit( self.image , self.rect )

class LifeContainer:
    
    def __init__( self , maxlife ):
        self.maxlife = maxlife
        self.life = maxlife
        self.image = alloc['image']['life_container']
        self.unit = alloc['image']['life_unit']
        self.unit_rect = self.unit.get_rect()
        self.pos = self.image.get_rect()
        self.rect = self.pos.move( ( 10 , alloc['screen_rect'].bottom - self.pos.bottom - 10 ) )

    def update( self ):
        alloc['screen'].blit( self.image , self.rect )
        if not self.life <= 0:
            nextPos = pygame.Rect( ( self.rect.left + self.unit_rect.right - 4 , self.rect.top + 8 ) , self.unit_rect.bottomright )
            units = int( self.life * ( 16.0 / self.maxlife ) ) + 1
            for x in range( units ):
                alloc['screen'].blit( self.unit , nextPos )
                nextPos.left += self.unit_rect.right + 4

class Hero:
    
    def __init__( self , initPos ):
        self.image = alloc['image']['hero'] 
        self.rect = self.image.get_rect().move( initPos )
        self.speed = 2
        self.life = LifeContainer( 50 )

    def _blit( self ):
        alloc['screen'].blit( self.image , self.rect )

    def _checkCollisions( self ):
        if self.life:
            alloc['shadows'][ tuple( alloc['tree'].shadow_rect ) ] = 1
            alloc['shadows'][ tuple( alloc['tree'].rect ) ] = 1
            collisions = self.rect.collidedict( alloc['shadows'] )
            alloc['shadows'].clear()
            if not collisions:
                self.life.life -= 1

    def _move( self , direction ):
        for x in range( self.speed ):
            self._step( direction )

    def _step( self , direction ):
        if direction == DIR_N and self.rect.top > alloc['ground_rect'].top:
            self.rect.top -= 1
        elif direction == DIR_E and self.rect.right < alloc['ground_rect'].right:
            self.rect.right += 1
        elif direction == DIR_S and self.rect.bottom < alloc['ground_rect'].bottom:
            self.rect.bottom += 1
        elif direction == DIR_W and self.rect.left > alloc['ground_rect'].left:
            self.rect.left -= 1
        else:
            return False
        return True

    def update( self , pressed ):
        # check shadow collisions
        self._checkCollisions()
        if not self.life.life:
            pass # the gnome is dead
        else:
            up, down, left, right = pressed
            # move based on keys
            if up:
                self._move( DIR_N )
            elif down:
                self._move( DIR_S )
            elif left:
                self._move( DIR_W )
            elif right:
                self._move( DIR_E )
        # blit
        self._blit()
        self.life.update()

####################
##                  
##  LEVEL CREATION! 
##                  

class Level:
    
    def __init__( self , level_number , caption , cloud_minspeed , cloud_maxspeed , time ):
        # Text
        self.level = alloc['font_large'].render( "Level " + str( level_number ) , True , ( 255 , 255 , 255 ) )
        self.caption = alloc['font_medium'].render( caption , True , ( 255 , 255 , 255 ) )
        self.cloud_minspeed = cloud_minspeed
        self.cloud_maxspeed = cloud_maxspeed
        self.time = time

    def _blitBackground( self ):
        # A little helper function for my background
        alloc['screen'].blit( alloc['image']['background'] , ( 0 , 0 ) )
        alloc['screen'].fill( ( 139 , 223 , 175 ) , alloc['ground_rect'] )

    def _levelScreen( self ):
        alloc['screen'].fill( ( 0 , 0 , 0 ) )
        alloc['screen'].blit( self.level , centerImgOnScreen( self.level , self.level.get_rect() ) )
        level_rect = self.level.get_rect()
        alloc['screen'].blit( self.caption , ( centerImgOnScreen( self.caption , self.caption.get_rect() )[0] , level_rect.bottom + level_rect.top + 5 ) )

    def _deathScreen( self ):
        alloc['screen'].fill( ( 0 , 0 , 0 ) )
        alloc['screen'].blit( alloc['death'] , centerImgOnScreen( alloc['death'] , alloc['death_rect'] ) )

    def _successScreen( self ):
        alloc['screen'].fill( ( 0 , 0 , 0 ) )
        alloc['screen'].blit( alloc['success'] , centerImgOnScreen( alloc['success'] , alloc['success_rect'] ) )
        alloc['screen'].blit( alloc['success_caption'] , ( centerImgOnScreen( alloc['success_caption'] , alloc['success_caption_rect'] )[0] , alloc['success_rect'].bottom + alloc['success_rect'].top + 5 ) )

    def _instantiate( self ):
        alloc['shadows'] = {}
        alloc['tree'] = Tree()
        alloc['clouds'] = Clouds( self.cloud_minspeed , self.cloud_maxspeed )
        alloc['hero'] = Hero( alloc['tree'].shadow_rect.topleft )

    def _eventLoop( self ):
        self._levelScreen()
        pygame.display.update()
        pygame.time.delay( 3000 )
        while True:
            # Blits and Updates
            self._blitBackground()
            # Check success/loss statuses
            if alloc['hero'].life.life <= 0:
                self._deathScreen()
                pygame.display.update()
                pygame.time.delay( 3000 )
                return LP_DEAD
            elif alloc['hero'].rect.colliderect( alloc['win_rect'] ):
                self._successScreen()
                pygame.display.update()
                pygame.time.delay( 3000 )
                return LP_WON
            pressed = pygame.key.get_pressed()
            ctrls = pressed[ pygame.K_w ], pressed[ pygame.K_s ], pressed[ pygame.K_a ], pressed[ pygame.K_d ]
            alloc['hero'].update( ctrls )
            alloc['clouds'].next()
            alloc['tree'].blit()
            # Event Loop
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    raise SystemExit
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_SPACE:
                        # PAUSE THE GAME
                        pause_center = ( alloc['screen_rect'].right / 2 ) - ( alloc['pause_rect'].right / 2 ) , ( alloc['screen_rect'].bottom / 2 ) - ( alloc['pause_rect'].bottom / 2 )
                        pause_topleft = alloc['screen_rect'].left + 10, alloc['screen_rect'].top + 10
                        alloc['screen'].blit( alloc['pause'] , pause_topleft )
                        pygame.display.update()
                        paused = True
                        while paused:
                            for event in pygame.event.get():
                                if event.type == pygame.QUIT:
                                    raise SystemExit
                                if event.type == pygame.KEYDOWN and event.key == pygame.K_SPACE:
                                    paused = False
                                    break
            # Update the screen
            pygame.display.update()

    def start( self ):
        self._instantiate()
        return self._eventLoop()

LEVELS = 6
alloc['level'] = range( LEVELS )
alloc['level'][0] = Level( 1 , "This should be easy..." , 1 , 1 , 60 )
alloc['level'][1] = Level( 2 , "A little harder..." , 1 , 2 , 60 )
alloc['level'][2] = Level( 3 , "Getting there..." , 2 , 2 , 60 )
alloc['level'][3] = Level( 4 , "Keep it up..." , 1 , 3 , 60 )
alloc['level'][4] = Level( 5 , "It's getting intense..." , 2 , 3 , 60 )
alloc['level'][5] = Level( 6 , "You're doing quite well..." , 3 , 3 , 60 )

#################
##               
##  TITLE SCREEN 
##               

class TitleScreen:

    def __init__( self ):
        self.clouds

#################
##               
##  DO THE LOOP! 
##               

class Driver:

    def _play( self ):
        for x in range( LEVELS ) :
            result = alloc['level'][ x ].start()
            if result == LP_DEAD:
                return False
            elif result == LP_WON and x == 5:
                return False
        return True

    def start( self ):
        # TODO: do some title screen stuff
        self._play()



###########################
##                         
##  Let's start this party 
##                         

if __name__ == '__main__':
    game = Driver()
    game.start()